# Copyright (C) 2007, 2008, 2009 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Functional tests for InteProxy"""

import sys
import os
import httplib
import unittest
import base64

import serversupport
import support

from inteproxy.transcoder import create_transcoder_map
from inteproxy.proxycore import MasterWorkerServer, InteProxyHTTPRequestHandler
from inteproxy.config import Struct, parse_inteproxy_url
import inteproxy.httpserver as httpserver
import inteproxy.getpassword


class ServerTest(unittest.TestCase, support.FileTestMixin):

    """Base class for tests that run the InteProxy in a thread"""

    ssl_key_and_cert = """\
-----BEGIN RSA PRIVATE KEY-----
MIICXAIBAAKBgQDHZXrYyr6lMox1DzapO+mFJ253e2epjIxTNuxLL/ibukO58ume
RbyVwNtZoSs9fmo69HCyaVIu4GRLBLYZsz+raKjpMv2FlQK5e7UJsj3dBYtGdCns
MWQ5qlqHJP9RGvxEsj707dLz9VpMYQ3Z+uwiXkcpQ0TAfuiNPjOrL2i1GwIDAQAB
AoGAEqjffywx/Rmm5UmALwQpTUmsX8gPomFBvIPWRw79L+2VavBdhm82P7lljZS0
jDOpU00OwtNyWmqrWA3f0LXijkhep7amwvoPr31I9PVxk2q8KlkqR854Kdo2T0ID
vvRJmsdjQSoraSNGwGhdRB3XXlE7mRrPUO6VyYQTMXY3SxECQQD7BJVcHnNLBRSA
/SOgpuWfp+uPJ+vmZX+kXqzaxB4Jdu2B4+4Ul2N2VB7lNA6zo38A2H2vGQfA88HU
yCuGU9LpAkEAy1qb1Jj2Hvgx8PlWy6aKsL90N8E4PKMh8ZtzwnJB9/5Ym8E6dwhr
G+SWyuF/Ay7rN/O6S+YblvtCoYhRz7PdYwJBAPSaNMk9Su/RlXdUNQF5YMzBHsOz
DSbHxSfwsdPDw9lJMIugBgG+u8c0lZ6XqbPXIA086MxVQb7+SOUF4ZPV3vECQBo8
I1SHM2GFdbP4BwmY/9WTraOvytiP10Y7XvDcGsSqzLWzdR58OI1NrWKZOvCnMfNy
/zrhgfe1jMAYPS3Fr68CQFKSaCQazOUnERcIaWK1/UcuBgMwmShcnENTHjQ8EBt1
ZYnyxKia81PR6W6xYSj/QqYqoC5uoXDKdAO2jYog1yQ=
-----END RSA PRIVATE KEY-----
-----BEGIN CERTIFICATE-----
MIIDLjCCApegAwIBAgIJAKvA4hSZg7hjMA0GCSqGSIb3DQEBBQUAMG4xCzAJBgNV
BAYTAkRFMRMwEQYDVQQIEwpTb21lLVN0YXRlMRgwFgYDVQQKEw9JbnRldmF0aW9u
IEdtYkgxHDAaBgNVBAsTE0ludGVQcm94eSBUZXN0c3VpdGUxEjAQBgNVBAMTCWxv
Y2FsaG9zdDAeFw0wOTAzMzExODI4MDFaFw0xOTAzMjkxODI4MDFaMG4xCzAJBgNV
BAYTAkRFMRMwEQYDVQQIEwpTb21lLVN0YXRlMRgwFgYDVQQKEw9JbnRldmF0aW9u
IEdtYkgxHDAaBgNVBAsTE0ludGVQcm94eSBUZXN0c3VpdGUxEjAQBgNVBAMTCWxv
Y2FsaG9zdDCBnzANBgkqhkiG9w0BAQEFAAOBjQAwgYkCgYEAx2V62Mq+pTKMdQ82
qTvphSdud3tnqYyMUzbsSy/4m7pDufLpnkW8lcDbWaErPX5qOvRwsmlSLuBkSwS2
GbM/q2io6TL9hZUCuXu1CbI93QWLRnQp7DFkOapahyT/URr8RLI+9O3S8/VaTGEN
2frsIl5HKUNEwH7ojT4zqy9otRsCAwEAAaOB0zCB0DAdBgNVHQ4EFgQUiX30Ifsl
A022O9V/JqUIMFVaAnUwgaAGA1UdIwSBmDCBlYAUiX30IfslA022O9V/JqUIMFVa
AnWhcqRwMG4xCzAJBgNVBAYTAkRFMRMwEQYDVQQIEwpTb21lLVN0YXRlMRgwFgYD
VQQKEw9JbnRldmF0aW9uIEdtYkgxHDAaBgNVBAsTE0ludGVQcm94eSBUZXN0c3Vp
dGUxEjAQBgNVBAMTCWxvY2FsaG9zdIIJAKvA4hSZg7hjMAwGA1UdEwQFMAMBAf8w
DQYJKoZIhvcNAQEFBQADgYEAgqZVnsRy5uM9bSlSzJPQ7s4AM2RqCnOYpqRMVzpM
aP4hEzoxht9VdNYbL3dW1WvL14srBcT+9B/mokFM5pDT/tevNbAFDvokh1VQNhmV
A5PtuG8pfsswvIl4+weagZgUjlxB8RGS0qFiuQqKbPnMiiRdMiJyhCHi67pA7foj
hnY=
-----END CERTIFICATE-----
"""


    # The content of the remote server in the format used by
    # serversupport.HTTPServer.  Override in derived classes.
    remote_contents = []

    # transcoders to use specified as a list of inteproxy urls
    transcoder_definitions = []

    # whether the server should rewrite urls
    rewrite_urls = False

    def setUp(self):
        self.ssl_cert = self.create_temp_file("cert.pem", self.ssl_key_and_cert)
        self.old_stderr = sys.stderr
        sys.stderr = open(os.path.devnull, "w")

        remote_server = self.create_origin_server()
        self.remote_server = httpserver.ServerThread(remote_server)
        self.remote_server.start(daemon=True)
        self.remote_server_base_url = "http://localhost:%d/" \
                                      % (self.remote_server.server_port,)

        http_proxy, username, password = self.create_http_proxy()
        if http_proxy:
            self.http_proxy = httpserver.ServerThread(http_proxy)
            self.http_proxy.start(daemon=True)
            self.http_proxy_desc = Struct(host="localhost",
                                          port=self.http_proxy.server_port,
                                          username=username, password=password)
        else:
            self.http_proxy = self.http_proxy_desc = None

        https_proxy, username, password = self.create_https_proxy()
        if https_proxy:
            self.https_proxy = httpserver.ServerThread(https_proxy)
            self.https_proxy.start(daemon=True)
            self.https_proxy_desc = Struct(host="localhost",
                                          port=self.https_proxy.server_port,
                                          username=username, password=password)
        else:
            self.https_proxy = self.https_proxy_desc = None


        proxyserver = MasterWorkerServer(("127.0.0.1", 0),
                                         InteProxyHTTPRequestHandler, 5,
                                         self.http_proxy_desc,
                                         self.https_proxy_desc,
                                   transcoder_map=self.create_transcoder_map(),
                                         rewrite_urls=self.rewrite_urls)
        self.server = httpserver.ServerThread(proxyserver)
        self.server.start(daemon=True)

    def create_transcoder_map(self):
        transcoder_map = transcoder_map=create_transcoder_map()
        for url in self.transcoder_definitions:
            transcoder_map.add_rule(parse_inteproxy_url(url))
        return transcoder_map

    def create_http_proxy(self):
        """Override in derived classes to create a proxy upstream from inteproxy

        The method should return a tuple with the proxy server object
        and the username and password for authentications against the
        proxy.
        """
        return None, None, None

    def create_https_proxy(self):
        """Override in derived classes to create an upstream https proxy

        The method should return a tuple with the proxy server object
        and the username and password for authentications against the
        proxy.
        """
        return None, None, None

    def create_origin_server(self):
        return serversupport.HTTPServer(self.remote_contents)

    def tearDown(self):
        try:
            for server in [self.remote_server,
                           self.http_proxy, self.https_proxy,
                           self.server]:
                if server is not None:
                    server.stop()
        finally:
            sys.stderr = self.old_stderr



class TestInteProxy(ServerTest):

    remote_contents = [
        ("/wms", [("Content-Type", "text/plain")], "wms data"),
        ("/search?q=foo#section1", [("Content-Type", "text/plain")],
         "some text"),
        ]

    def test_httpproxy_passthrough(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET", self.remote_server_base_url + "wms")
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data, "wms data")

    def test_httpproxy_with_query_and_fragment(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET",
                     self.remote_server_base_url + "search?q=foo#section1")
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data, "some text")


class TestInteProxyOWSProxy(ServerTest):

    remote_contents = [
        ("/wms?user=fred&password=12345",
         [("Content-Type", "text/plain")], "wms data"),
        ("/search?q=foo&user=fred&password=12345#section1",
         [("Content-Type", "text/plain")], "some text"),
        ]

    def create_origin_server(self):
        return serversupport.HTTPSServer(self.remote_contents,
                                         certfile=self.ssl_cert)

    def create_transcoder_map(self):
        url_prefix = ("owsproxy://fred:12345@localhost:%d"
                      % self.remote_server.server_port)
        transcoder_map = create_transcoder_map()
        transcoder_map.add_rule(parse_inteproxy_url(url_prefix + "/wms"))
        transcoder_map.add_rule(parse_inteproxy_url(url_prefix + "/search"))
        return transcoder_map

    if serversupport.HTTPSServer.supported:
        def test_httpproxy_passthrough(self):
            http = httplib.HTTPConnection("localhost", self.server.server_port)
            http.request("GET", self.remote_server_base_url + "wms")
            response = http.getresponse()
            self.assertEquals(response.status, 200)
            data = response.read()
            self.assertEquals(data, "wms data")

        def test_httpproxy_with_query_and_fragment(self):
            http = httplib.HTTPConnection("localhost", self.server.server_port)
            http.request("GET",
                         self.remote_server_base_url + "search?q=foo#section1")
            response = http.getresponse()
            self.assertEquals(response.status, 200)
            data = response.read()
            self.assertEquals(data, "some text")


class TestInteProxyURLRewriting(ServerTest):

    remote_contents = [
        ("/rewrite-01", [("Content-Type", "text/plain")],
         "An url that may be rewritten: http://example.com/a/wms"),
        ("/rewrite-02", [("Content-Type", "text/plain")],
         "An URL that may be rewritten: http://example.com/a/wms\n"
         "and one that may not: http://example.com/foo\n"
         "and one that may: http://example.com/another/wms\n"),
        ("/rewrite-03", [("Content-Type", "text/plain")],
         "An URL that may be rewritten: https://example.com/a/wms\n"
         "and one that may not: http://example.com/foo\n"
         "and one that may: http://example.com/another/wms\n"),
        ("/rewrite-04", [("Content-Type", "text/plain")],
         "An URL that may be rewritten: https://example.com/pattern/wms\n"
         "and one that may not: http://example.com/wms\n"
         "and one that may: http://example.com/pattern/bar/baz\n"),
        ("/rewrite-05", [("Content-Type", "text/plain")],
         "An URL that may be rewritten: https://data.intevation.de/wms\n"
         "and one that may not: http://data.intevation.de/gis/wms\n"
         "and one that may: http://frida.intevation.org/wms/simple\n"),
        ("/rewrite-06", [("Content-Type", "text/plain")],
         "An long line: an URL that may be rewritten:"
         " https://data.intevation.de/wms and one that may not:"
         " http://data.intevation.de/gis/wms and another one that may:"
         " http://frida.intevation.org/wms/simple\n"),
        ]


    transcoder_definitions = ["owsproxy://example.com/a/wms",
                              "owsproxy://example.com/another/wms",
                              "owsproxy://example.com/pattern/*",
                              "owsproxy://*.intevation.*/wms*",
                              ]

    rewrite_urls = True

    def test_httpproxy_url_rewriting_single_url(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET",
                     "http://localhost:%d/localhost:%d/rewrite-01"
                     % (self.server.server_port,
                        self.remote_server.server_port))
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data,
                          "An url that may be rewritten:"
                          " http://localhost:%d/example.com/a/wms"
                          % (self.server.server_port,))

    def test_httpproxy_url_rewriting_multiple_urls(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET",
                     "http://localhost:%d/localhost:%d/rewrite-02"
                     % (self.server.server_port,
                        self.remote_server.server_port))
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data,
                          "An URL that may be rewritten:"
                          " http://localhost:%d/example.com/a/wms\n"
                          "and one that may not: http://example.com/foo\n"
                          "and one that may:"
                          " http://localhost:%d/example.com/another/wms\n"
                          % (self.server.server_port,
                             self.server.server_port))

    def test_httpproxy_url_rewriting_https_urls(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET",
                     "http://localhost:%d/localhost:%d/rewrite-03"
                     % (self.server.server_port,
                        self.remote_server.server_port))
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data,
                          "An URL that may be rewritten:"
                          " http://localhost:%d/example.com/a/wms\n"
                          "and one that may not: http://example.com/foo\n"
                          "and one that may:"
                          " http://localhost:%d/example.com/another/wms\n"
                          % (self.server.server_port,
                             self.server.server_port))

    def test_httpproxy_url_rewriting_with_patterns(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET",
                     "http://localhost:%d/localhost:%d/rewrite-04"
                     % (self.server.server_port,
                        self.remote_server.server_port))
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data,
                          "An URL that may be rewritten:"
                          " http://localhost:%d/example.com/pattern/wms\n"
                          "and one that may not: http://example.com/wms\n"
                          "and one that may:"
                          " http://localhost:%d/example.com/pattern/bar/baz\n"
                          % (self.server.server_port,
                             self.server.server_port))

    def test_httpproxy_url_rewriting_with_host_patterns(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET",
                     "http://localhost:%d/localhost:%d/rewrite-05"
                     % (self.server.server_port,
                        self.remote_server.server_port))
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data,
                          "An URL that may be rewritten:"
                   " http://localhost:%d/data.intevation.de/wms\n"
                   "and one that may not: http://data.intevation.de/gis/wms\n"
                   "and one that may:"
                   " http://localhost:%d/frida.intevation.org/wms/simple\n"
                          % (self.server.server_port,
                             self.server.server_port))
    def test_httpproxy_url_rewriting_with_long_lines(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET",
                     "http://localhost:%d/localhost:%d/rewrite-06"
                     % (self.server.server_port,
                        self.remote_server.server_port))
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data,
                          "An long line: an URL that may be rewritten:"
                          " http://localhost:%d/data.intevation.de/wms"
                          " and one that may not:"
                          " http://data.intevation.de/gis/wms"
                          " and another one that may:"
                     " http://localhost:%d/frida.intevation.org/wms/simple\n"
                          % (self.server.server_port,
                             self.server.server_port))


class TestInteProxyWithExtraProxy(ServerTest):

    remote_contents = [
        ("/wms", [("Content-Type", "text/plain")], "wms data"),
        ]

    def create_http_proxy(self):
        server = MasterWorkerServer(("127.0.0.1", 0),
                                    InteProxyHTTPRequestHandler, 5,
                                    None, None,
                                    transcoder_map=create_transcoder_map())
        return server, None, None

    def test(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET", self.remote_server_base_url + "wms")
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data, "wms data")


class AuthenticatingProxyRequestHandler(InteProxyHTTPRequestHandler):

    """Special request handler for remote proxy authorization tests

    This request handler only accepts requests that send a
    Proxy-Authorization header with Basic authentication scheme,
    username 'testuser' and password 'secret'
    """

    def handle_proxy_request(self):
        authenticated = False
        auth = self.headers.get("Proxy-Authorization")
        if auth:
            parts = auth.split()
            if len(parts) == 2:
                auth_scheme, encoded = parts
                if auth_scheme.lower() == "basic":
                    decoded = base64.b64decode(encoded)
                    if decoded == "testuser:secret":
                        authenticated = True

        if not authenticated:
            # we should send a Proxy-Authenticate too, but it doesn't
            # matter for this test because InteProxy doesn't look at the
            # header yet.
            self.send_response(407, "Proxy Authentication Required")
            self.send_header("Proxy-Authenticate",
                             'Basic realm="inteproxytest"')
            self.send_header("Content-Type", "text/html")
            self.send_header('Connection', 'close')
            self.end_headers()
            self.wfile.write("<html><body>"
                             "Proxy requires authentication"
                             "</body><html>")

        else:
            InteProxyHTTPRequestHandler.handle_proxy_request(self)


class TestInteProxyWithExtraProxyWithAuthentication(ServerTest):

    remote_contents = [
        ("/wms", [("Content-Type", "text/plain")], "wms data"),
        ]

    def create_http_proxy(self):
        server = MasterWorkerServer(("127.0.0.1", 0),
                                    AuthenticatingProxyRequestHandler, 5,
                                    None, None,
                                    transcoder_map=create_transcoder_map())
        return server, "testuser", "secret"

    def test(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET", self.remote_server_base_url + "wms")
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data, "wms data")


class TestInteProxyWithExtraProxyWithIncorrectAuthentication(ServerTest):

    remote_contents = [
        ("/wms", [("Content-Type", "text/plain")], "wms data"),
        ]

    def create_http_proxy(self):
        server = MasterWorkerServer(("127.0.0.1", 0),
                                    AuthenticatingProxyRequestHandler, 5,
                                    None, None,
                                    transcoder_map=create_transcoder_map())
        return server, "testuser", "wrong"

    def test(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET", self.remote_server_base_url + "wms")
        response = http.getresponse()
        # the remote proxy returns a 407 (Proxy Authentication
        # Required), but it must not be returned to the inteproxy client
        # as such, otherwise inteproxy's client would believe that
        # inteproxy itself required authentication.  Inteproxy should
        # send a 502 (Bad Gateway) instead to indicate a server
        # configuration problem
        self.assertEquals(response.status, 502)

        # there must not be a Proxy-Authenticate header either
        self.failIf("Proxy-Authenticate" in response.msg, None)


class TestInteProxyOWSProxyWithAuthenticatingHTTPSProxy(ServerTest):

    remote_contents = [
        ("/wms?user=fred&password=12345",
         [("Content-Type", "text/plain")], "wms data"),
        ("/search?q=foo&user=fred&password=12345#section1",
         [("Content-Type", "text/plain")], "some text"),
        ]

    def create_https_proxy(self):
        server = MasterWorkerServer(("127.0.0.1", 0),
                                    serversupport.CONNECTRequestHandler, 5,
                                    None, None,
                                    transcoder_map=create_transcoder_map())
        return server, "testuser", "secret"

    def create_origin_server(self):
        return serversupport.HTTPSServer(self.remote_contents,
                                         certfile=self.ssl_cert)

    def create_transcoder_map(self):
        url_prefix = ("owsproxy://fred:12345@localhost:%d"
                      % self.remote_server.server_port)
        transcoder_map = create_transcoder_map()
        transcoder_map.add_rule(parse_inteproxy_url(url_prefix + "/wms"))
        transcoder_map.add_rule(parse_inteproxy_url(url_prefix + "/search"))
        return transcoder_map

    if serversupport.HTTPSServer.supported:
        def test_httpproxy_passthrough(self):
            http = httplib.HTTPConnection("localhost", self.server.server_port)
            http.request("GET", self.remote_server_base_url + "wms")
            response = http.getresponse()
            self.assertEquals(response.status, 200)
            data = response.read()
            self.assertEquals(data, "wms data")

        def test_httpproxy_with_query_and_fragment(self):
            http = httplib.HTTPConnection("localhost", self.server.server_port)
            http.request("GET",
                         self.remote_server_base_url + "search?q=foo#section1")
            response = http.getresponse()
            self.assertEquals(response.status, 200)
            data = response.read()
            self.assertEquals(data, "some text")
