# Copyright (C) 2007, 2008, 2009 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Tests for TranscoderMap"""

import unittest

import support

from inteproxy.transcoder import create_transcoder_map, \
     OWSProxyGETTranscoder, OWSProxyPOSTTranscoder, BasicAuthTranscoder, \
     IdentityTranscoder
import inteproxy.config

class TranscoderMapTest(unittest.TestCase, support.FileTestMixin):

    config_contents = ""

    def setUp(self):
        config_filename = self.create_temp_file(self.id(), self.config_contents)
        config = inteproxy.config.read_config(config_filename)
        self.transcoder_map = create_transcoder_map()
        self.transcoder_map.add_rules(config.rules)


class TestTranscoderMapNoWildcards(TranscoderMapTest):

    config_contents = """\
[inteproxy-rules]
urls=owsproxy://inteproxy-demo.intevation.org/cgi-bin/frida-wms
     basicauth://example.com/foo
"""
    #

    def test_get_transcoder_http_proxy(self):
        """Test the get_transcoder method in http-proxy mode"""
        test_cases = [
            ("GET", "http://inteproxy-demo.intevation.org/cgi-bin/frida-wms",
             OWSProxyGETTranscoder),
            ("POST", "http://inteproxy-demo.intevation.org/cgi-bin/frida-wms",
             OWSProxyPOSTTranscoder),
            ("GET", "http://inteproxy-demo.intevation.org/cgi-bin/frida",
             IdentityTranscoder),
            ("POST", "http://inteproxy-demo.intevation.org/cgi-bin/frida",
             IdentityTranscoder),

            ("GET", "http://example.com/foo", BasicAuthTranscoder),
            ("POST", "http://example.com/foo", BasicAuthTranscoder),

            # Example for a non-WMS request sent by deejump
            ("GET",
        "http://schemas.opengeospatial.net/wms/1.1.1/WMS_MS_Capabilities.dtd",
             IdentityTranscoder),
            ]

        for method, url, expectedcls in test_cases:
            transcoder = self.transcoder_map.get_transcoder(method, url)
            cls = transcoder.__class__
            msg=("Wrong class for %(method)s %(url)s:"
                 " got %(cls)s expected %(expectedcls)s" % locals())
            self.assertEquals(cls, expectedcls, msg)

    def test_get_transcoder_inteproxy_style(self):
        """Test the get_transcoder method in InteProxy mode"""
        test_cases = [
            ("GET", "/inteproxy-demo.intevation.org/cgi-bin/frida-wms",
             OWSProxyGETTranscoder),
            ("POST", "/inteproxy-demo.intevation.org/cgi-bin/frida-wms",
             OWSProxyPOSTTranscoder),
            ("GET", "/inteproxy-demo.intevation.org/cgi-bin/frida",
             IdentityTranscoder),
            ("POST", "/inteproxy-demo.intevation.org/cgi-bin/frida",
             IdentityTranscoder),

            ("GET", "/example.com/foo", BasicAuthTranscoder),
            ("POST", "/example.com/foo", BasicAuthTranscoder),
            ]

        for method, url, expectedcls in test_cases:
            transcoder = self.transcoder_map.get_transcoder(method, url)
            cls = transcoder.__class__
            msg=("Wrong class for %(method)s %(url)s:"
                 " got %(cls)s expected %(expectedcls)s" % locals())
            self.assertEquals(cls, expectedcls, msg)


class TestTranscoderMapWithPathWildcards(TranscoderMapTest):

    config_contents = """\
[inteproxy-rules]
urls=owsproxy://inteproxy-demo.intevation.org/cgi-bin/*
     basicauth://intevation.de/prefix*suffix
"""
    #

    def test_get_transcoder_http_proxy(self):
        """Test the get_transcoder method in http-proxy mode with wildcards"""
        test_cases = [
            ("GET", "http://inteproxy-demo.intevation.org/cgi-bin/frida-wms",
             OWSProxyGETTranscoder),
            ("POST", "http://inteproxy-demo.intevation.org/cgi-bin/frida-wms",
             OWSProxyPOSTTranscoder),
            ("GET", "http://inteproxy-demo.intevation.org/otherdir",
             IdentityTranscoder),
            ("POST", "http://inteproxy-demo.intevation.org/otherdir",
             IdentityTranscoder),

            ("GET", "http://intevation.de/prefix/bla/blub/suffix",
             BasicAuthTranscoder),
            ("POST", "http://intevation.de/prefix/bla/blub/suffix",
             BasicAuthTranscoder),

            # not matched:
            ("GET", "http://intevation.de/prefix/bla/", IdentityTranscoder),
            ("POST", "http://intevation.de/prefix/bla/", IdentityTranscoder),
            ("GET", "http://intevation.de/prefix/bla/suffix/",
             IdentityTranscoder),
            ("POST", "http://intevation.de/prefix/bla/suffix/more",
             IdentityTranscoder),
            ]

        for method, url, expectedcls in test_cases:
            transcoder = self.transcoder_map.get_transcoder(method, url)
            cls = transcoder.__class__
            msg=("Wrong class for %(method)s %(url)s:"
                 " got %(cls)s expected %(expectedcls)s" % locals())
            self.assertEquals(cls, expectedcls, msg)


class TestTranscoderMapWithHostWildcards(TranscoderMapTest):

    config_contents = """\
[inteproxy-rules]
urls=owsproxy://*intevation.*/wms/*
     basicauth://*.example.com/*
"""
    #

    def test_get_transcoder_http_proxy(self):
        """Test the get_transcoder method in http-proxy mode with wildcards"""
        test_cases = [
            # matched by the first rule
            ("GET", "http://intevation.de/wms/demo",
             OWSProxyGETTranscoder),
            ("POST", "http://intevation.org/wms/demo",
             OWSProxyPOSTTranscoder),
            ("GET", "http://gis.intevation.org/wms/bla/",
             OWSProxyGETTranscoder),
            ("POST", "http://data.intevation.de/wms/frida",
             OWSProxyPOSTTranscoder),

            # not matched by the first rule
            ("GET", "http://intevation.de/prefix/wms/demo",
             IdentityTranscoder),
            ("POST", "http://intevation.de/prefix/wms/demo",
             IdentityTranscoder),

            # matched by the second rule
            ("GET", "http://geodata.example.com/prefix/frida/wms",
             BasicAuthTranscoder),
            ("POST", "http://geodata.example.com/prefix/wms",
             BasicAuthTranscoder),
            ("GET", "http://frida.example.com/", BasicAuthTranscoder),
            ("POST", "http://gis.data.example.com/", BasicAuthTranscoder),

            # not matched by the second rule
            ("GET", "http://geodata.example.org/frida/wms",
             IdentityTranscoder),
            ("POST", "http://geodata.example.net/prefix/wms",
             IdentityTranscoder),
            ("GET", "http://frida.example.org/", IdentityTranscoder),
            ("POST", "http://data.example.net/", IdentityTranscoder),
            ("GET", "http://geodata.example.com.net/frida/wms",
             IdentityTranscoder),
            ]

        for method, url, expectedcls in test_cases:
            transcoder = self.transcoder_map.get_transcoder(method, url)
            cls = transcoder.__class__
            msg=("Wrong class for %(method)s %(url)s:"
                 " got %(cls)s expected %(expectedcls)s" % locals())
            self.assertEquals(cls, expectedcls, msg)


class TestTranscoderMapWithPorts(TranscoderMapTest):

    config_contents = """\
[inteproxy-rules]
urls=owsproxy://*intevation.*:8080/wms/
     basicauth://*.example.com:9000/*
     owsproxy://*.example.com/*
"""
    #

    def test_get_transcoder_http_proxy(self):
        """Test the get_transcoder method in http-proxy mode with wildcards"""
        test_cases = [
            # matched by the first rule
            ("GET", "http://intevation.de:8080/wms/", OWSProxyGETTranscoder),
            ("POST", "http://intevation.org:8080/wms/", OWSProxyPOSTTranscoder),
            ("GET", "http://gis.intevation.org:8080/wms/",
             OWSProxyGETTranscoder),
            ("POST", "http://data.intevation.de:8080/wms/",
             OWSProxyPOSTTranscoder),

            # not matched by the first rule
            ("GET", "http://intevation.de:9000/wms/", IdentityTranscoder),
            ("POST", "http://intevation.de//wms/", IdentityTranscoder),

            # matched by the second rule
            ("GET", "http://geodata.example.com:9000/prefix/frida/wms",
             BasicAuthTranscoder),
            ("POST", "http://geodata.example.com:9000/prefix/wms",
             BasicAuthTranscoder),
            ("GET", "http://frida.example.com:9000/", BasicAuthTranscoder),
            ("POST", "http://gis.data.example.com:9000/", BasicAuthTranscoder),

            # not matched by the second rule
            ("GET", "http://geodata.example.org:9001/frida/wms",
             IdentityTranscoder),
            ("POST", "http://geodata.example.net:8080/prefix/wms",
             IdentityTranscoder),

            # matched by the third rule
            ("GET", "http://geodata.example.com/prefix/frida/wms",
             OWSProxyGETTranscoder),
            ("POST", "http://geodata.example.com/prefix/wms",
             OWSProxyPOSTTranscoder),
            ]

        for method, url, expectedcls in test_cases:
            transcoder = self.transcoder_map.get_transcoder(method, url)
            cls = transcoder.__class__
            msg=("Wrong class for %(method)s %(url)s:"
                 " got %(cls)s expected %(expectedcls)s" % locals())
            self.assertEquals(cls, expectedcls, msg)
